"""
Tests for functionalities in geoseries_util.
"""

import geopandas as gpd
import pytest
import shapely
import shapely.geometry as sh_geom
from pygeoops import GeometryType

import geofileops as gfo
import geofileops._compat as compat
from geofileops.util import _geoseries_util
from tests import test_helper


def test_get_geometrytypes():
    test_gdf = gpd.GeoDataFrame(
        geometry=[
            None,
            sh_geom.Point(),
            sh_geom.LineString(),
            sh_geom.Polygon(),
            test_helper.TestData.point,
            test_helper.TestData.multipoint,
            test_helper.TestData.polygon_with_island,
            test_helper.TestData.polygon_no_islands,
            test_helper.TestData.multipolygon,
            test_helper.TestData.geometrycollection,
        ]
    )
    # None and empty geometries are by default ignored in get_geometrytypes
    test_geometrytypes = _geoseries_util.get_geometrytypes(test_gdf.geometry)
    assert len(test_geometrytypes) == 5

    test_gdf = gpd.GeoDataFrame(
        geometry=[
            None,
            sh_geom.Point(),
            sh_geom.LineString(),
            sh_geom.Polygon(),
            test_helper.TestData.point,
        ]
    )
    # None and empty geometries are by default ignored in get_geometrytypes
    test_geometrytypes = _geoseries_util.get_geometrytypes(test_gdf.geometry)
    assert len(test_geometrytypes) == 1
    assert GeometryType.POINT in test_geometrytypes

    # Empty geometries are counted with ignore_empty_geometries=False.
    test_gdf = gpd.GeoDataFrame(
        geometry=[
            None,
            sh_geom.Point(),
            sh_geom.LineString(),
            sh_geom.Polygon(),
            test_helper.TestData.point,
        ]
    )
    test_geometrytypes = _geoseries_util.get_geometrytypes(
        test_gdf.geometry, ignore_empty_geometries=False
    )
    # In shapely 2, empty geometries get the correct type, in shapely 1 they were always
    # of type geometrycollection
    if compat.SHAPELY_GTE_20:
        assert len(test_geometrytypes) == 3
        assert GeometryType.POINT in test_geometrytypes
        assert GeometryType.LINESTRING in test_geometrytypes
        assert GeometryType.POLYGON in test_geometrytypes
        assert GeometryType.GEOMETRYCOLLECTION not in test_geometrytypes
    else:
        assert len(test_geometrytypes) == 2


def test_harmonize_geometrytypes():
    # Test for gdf with None + point + multipoint -> all multipoint
    test_gdf = gpd.GeoDataFrame(
        geometry=[
            None,
            sh_geom.Point(),
            test_helper.TestData.point,
            test_helper.TestData.multipoint,
            test_helper.TestData.point,
            test_helper.TestData.multipoint,
        ]
    )
    test_gdf_geometrytypes = _geoseries_util.get_geometrytypes(test_gdf.geometry)
    assert len(test_gdf_geometrytypes) == 2
    test_result_gdf = test_gdf.copy()
    assert isinstance(test_result_gdf, gpd.GeoDataFrame)
    test_result_gdf.geometry = _geoseries_util.harmonize_geometrytypes(
        test_result_gdf.geometry
    )
    test_result_geometrytypes = _geoseries_util.get_geometrytypes(
        test_result_gdf.geometry
    )
    assert len(test_result_geometrytypes) == 1
    assert test_result_geometrytypes[0] == GeometryType.MULTIPOINT
    for index, geom in test_result_gdf.geometry.items():
        if index in [0, 1]:
            assert geom is None
        else:
            assert geom is not None

    # Test for gdf with linestring + multilinestring -> all multilinestring
    test_gdf = gpd.GeoDataFrame(
        geometry=[
            None,
            sh_geom.LineString(),
            test_helper.TestData.linestring,
            test_helper.TestData.multilinestring,
            test_helper.TestData.linestring,
            test_helper.TestData.multilinestring,
        ]
    )
    test_gdf_geometrytypes = _geoseries_util.get_geometrytypes(test_gdf.geometry)
    assert len(test_gdf_geometrytypes) == 2
    test_result_gdf = test_gdf.copy()
    assert isinstance(test_result_gdf, gpd.GeoDataFrame)
    test_result_gdf.geometry = _geoseries_util.harmonize_geometrytypes(
        test_result_gdf.geometry
    )
    test_result_geometrytypes = _geoseries_util.get_geometrytypes(
        test_result_gdf.geometry
    )
    assert len(test_result_geometrytypes) == 1
    assert test_result_geometrytypes[0] == GeometryType.MULTILINESTRING
    for index, geom in test_result_gdf.geometry.items():
        if index in [0, 1]:
            assert geom is None
        else:
            assert geom is not None

    # Test for gdf with polygon + multipolygon -> all multipolygon
    test_gdf = gpd.GeoDataFrame(
        geometry=[
            test_helper.TestData.polygon_with_island,
            None,
            sh_geom.Polygon(),
            test_helper.TestData.polygon_with_island,
            test_helper.TestData.multipolygon,
        ]
    )
    test_gdf_geometrytypes = _geoseries_util.get_geometrytypes(test_gdf.geometry)
    assert len(test_gdf_geometrytypes) == 2
    # Filter the gdf a bit to test that the indexes are retained properly in
    test_gdf = test_gdf.iloc[[1, 2, 3, 4]]
    test_result_gdf = test_gdf.copy()
    assert isinstance(test_result_gdf, gpd.GeoDataFrame)
    test_result_gdf.geometry = _geoseries_util.harmonize_geometrytypes(
        test_result_gdf.geometry
    )
    test_result_geometrytypes = _geoseries_util.get_geometrytypes(
        test_result_gdf.geometry
    )
    assert len(test_result_geometrytypes) == 1
    assert test_result_geometrytypes[0] == GeometryType.MULTIPOLYGON
    for index, geom in test_result_gdf.geometry.items():
        if index in [1, 2]:
            assert geom is None
        else:
            assert geom is not None

    # Test for gdf with all types of geometrytypes -> no harmonization possible
    test_gdf = gpd.GeoDataFrame(
        geometry=[
            None,
            sh_geom.Polygon(),
            test_helper.TestData.point,
            test_helper.TestData.multipoint,
            test_helper.TestData.polygon_with_island,
            test_helper.TestData.polygon_no_islands,
            test_helper.TestData.multipolygon,
            test_helper.TestData.geometrycollection,
        ]
    )
    test_geometrytypes = _geoseries_util.get_geometrytypes(test_gdf.geometry)
    assert len(test_geometrytypes) == 5
    test_result_gdf = test_gdf.copy()
    assert isinstance(test_result_gdf, gpd.GeoDataFrame)
    test_result_gdf.geometry = _geoseries_util.harmonize_geometrytypes(
        test_result_gdf.geometry
    )
    test_result_geometrytypes = _geoseries_util.get_geometrytypes(
        test_result_gdf.geometry
    )
    assert len(test_result_geometrytypes) == 5
    for index, geom in test_result_gdf.geometry.items():
        if index in [0]:
            # Only None is None, empty geometry is not changed!
            assert geom is None
        else:
            assert geom is not None


def test_is_valid_reason(tmp_path):
    # Test with valid data + Empty geometry
    # -------------------------------------
    test_gdf = gpd.GeoDataFrame(
        geometry=[
            sh_geom.Polygon(),
            test_helper.TestData.point,
            test_helper.TestData.multipoint,
            test_helper.TestData.polygon_with_island,
            test_helper.TestData.polygon_no_islands,
            test_helper.TestData.multipolygon,
            test_helper.TestData.geometrycollection,
        ]
    )
    result = _geoseries_util.is_valid_reason(test_gdf.geometry)

    assert len(result) == len(test_gdf)
    assert result.unique() == "Valid Geometry"

    # Test if indexes are retained
    # ----------------------------
    test_filtered_gdf = test_gdf[3:-1]
    assert isinstance(test_filtered_gdf.geometry, gpd.GeoSeries)
    result = _geoseries_util.is_valid_reason(test_filtered_gdf.geometry)

    assert len(result) == len(test_filtered_gdf)
    assert result.unique() == "Valid Geometry"
    assert result.index.to_list() == test_filtered_gdf.index.to_list()

    # Test with None
    # --------------
    test_gdf = gpd.GeoDataFrame(geometry=[None])
    result = _geoseries_util.is_valid_reason(test_gdf.geometry)

    # is_valid_reason returns None for None geometries
    assert result[0] is None

    # Test with invalid data
    # ----------------------
    # Prepare test data
    path = test_helper.get_testfile("polygon-invalid")
    gdf = gfo.read_file(path)
    result = _geoseries_util.is_valid_reason(gdf.geometry)

    assert len(result) == len(gdf)
    assert result[0].startswith("Ring Self-intersection")


poly_gridsize_error = shapely.from_wkt(
    "Polygon ((26352.5 175096.6, 26352.6 175096.6, 26352.6 175096.7, "
    "26352.5 175096.7, 26352.5 175096.6),(26352.528000000002 175096.676, "
    "26352.528369565214 175096.67489130437, 26352.528140495866 175096.67619834712, "
    "26352.52785714286 175096.67714285714, 26352.53 175096.66, "
    "26352.528000000002 175096.676))"
)
poly_ok = shapely.from_wkt(
    "Polygon ((26352.5 175096.7, 26352.66895 175096.76895, 26352.6 175096.6, "
    "26352.5 175096.6, 26352.5 175096.7))"
)


@pytest.mark.parametrize("raise_on_topoerror", [True, False])
@pytest.mark.parametrize(
    "geometry, exp_geometry",
    [
        (
            [poly_gridsize_error, poly_ok],
            [poly_gridsize_error, shapely.set_precision(poly_ok, grid_size=0.01)],
        ),
        (poly_gridsize_error, poly_gridsize_error),
    ],
)
def test_set_precision(geometry, exp_geometry, raise_on_topoerror):
    # The only currently known test case only works with geos 3.12
    if shapely.geos_version != (3, 12, 0):
        pytest.skip()

    grid_size = 0.01

    if raise_on_topoerror:
        with pytest.raises(Exception, match="TopologyException: Ring edge missing at"):
            # raise_on_topoerror is default False
            result = _geoseries_util.set_precision(geometry, grid_size=grid_size)
        return

    result = _geoseries_util.set_precision(
        geometry, grid_size=grid_size, raise_on_topoerror=raise_on_topoerror
    )

    # Check result
    if hasattr(geometry, "__len__"):
        assert result.tolist() == exp_geometry
    else:
        assert result == exp_geometry



"""
Tests for functionalities in helpers.layer_styles.
"""

import pytest

import geofileops as gfo
from geofileops.helpers import layerstyles
from tests import test_helper


def test_has_layerstyles_table(tmp_path):
    test_path = test_helper.get_testfile("polygon-parcel", dst_dir=tmp_path)

    assert not layerstyles._has_layerstyles_table(test_path)
    layerstyles._init_layerstyles(test_path)
    assert layerstyles._has_layerstyles_table(test_path)


def test_add_get_remove_layer_styles(tmp_path):
    test_path = test_helper.get_testfile("polygon-parcel", dst_dir=tmp_path)

    assert not layerstyles._has_layerstyles_table(test_path)
    layerstyles._init_layerstyles(test_path)
    assert layerstyles._has_layerstyles_table(test_path)

    # Add layer style to "parcel" layer
    with open(test_helper._data_dir / "polygonstyle.sld") as file:
        sld = file.read()
    with open(test_helper._data_dir / "polygonstyle.qml") as file:
        qml = file.read()
    gfo.add_layerstyle(
        path=test_path,
        layer="parcels",
        name="test_style",
        sld=sld,
        qml=qml,
        use_as_default=True,
    )
    layerstyles_df = gfo.get_layerstyles(test_path)
    assert len(layerstyles_df) == 1

    # Adding the same style again should give an error
    with pytest.raises(ValueError, match="layer style exists already: "):
        gfo.add_layerstyle(
            path=test_path, layer="parcels", name="test_style", qml="test_qml"
        )

    # Backup the styled file to be a able to check it manually in QGIS
    styled_path = tmp_path / f"{test_path.stem}_styled{test_path.suffix}"
    gfo.copy(src=test_path, dst=styled_path)

    # Remove the style again
    gfo.remove_layerstyle(test_path, id=1)
    layerstyles_df = gfo.get_layerstyles(test_path)
    assert len(layerstyles_df) == 0

    # Removing a style that doesn't exist is OK
    gfo.remove_layerstyle(test_path, id=1)



"""
Tests for operations that are executed using a sql statement on one layer.
"""

import logging
import math
from importlib import import_module
from typing import Any, Optional

import geopandas as gpd
import pytest
import shapely
from shapely import MultiPolygon, Polygon

from geofileops import GeometryType, fileops, geoops
from geofileops._compat import SPATIALITE_GTE_51
from geofileops.util import _geofileinfo, _geoops_sql
from geofileops.util import _io_util as io_util
from geofileops.util._geofileinfo import GeofileInfo
from tests import test_helper
from tests.test_helper import (
    EPSGS,
    GRIDSIZE_DEFAULT,
    SUFFIXES_GEOOPS,
    TESTFILES,
    WHERE_AREA_GT_400,
    assert_geodataframe_equal,
)

# Init gfo module
current_geoops_module = "unknown"
GEOOPS_MODULES = [
    "geofileops.geoops",
    "geofileops.util._geoops_gpd",
    "geofileops.util._geoops_sql",
]


def set_geoops_module(geoops_module: str):
    global current_geoops_module
    if current_geoops_module == geoops_module:
        # The right module is already loaded, so don't do anything
        return
    else:
        # Load the desired module as fileops
        global geoops
        geoops = import_module(geoops_module, __package__)
        current_geoops_module = geoops_module
        print(f"gfo module switched to: {current_geoops_module}")


def basic_combinations_to_test(
    geoops_modules: list[str] = GEOOPS_MODULES,
    testfiles: list[str] = TESTFILES,
    epsgs: list[int] = EPSGS,
    suffixes: list[str] = SUFFIXES_GEOOPS,
) -> list[Any]:
    """
    Return sensible combinations of parameters to be used in tests for following params:
        suffix, epsg, geoops_module, testfile, empty_input, gridsize, where_post
    """
    result = []

    # On .gpkg test:
    #   - all combinations of geoops_modules, testfiles and epsgs
    #   - fixed empty_input, suffix
    #   - dimensions="XYZ" for polygon input
    for epsg in epsgs:
        for geoops_module in geoops_modules:
            for testfile in testfiles:
                where_post = None
                keep_empty_geoms: Optional[bool] = False
                dimensions = None
                gridsize = 0.01 if epsg == 31370 else GRIDSIZE_DEFAULT
                if testfile == "polygon-parcel":
                    dimensions = "XYZ"
                    keep_empty_geoms = None
                    if epsg == 31370:
                        where_post = WHERE_AREA_GT_400
                elif testfile == "point":
                    keep_empty_geoms = True
                result.append(
                    (
                        ".gpkg",
                        epsg,
                        geoops_module,
                        testfile,
                        False,
                        gridsize,
                        keep_empty_geoms,
                        where_post,
                        dimensions,
                    )
                )

    # On other suffixes test:
    #   - all combinations of geoops_modules, testfiles
    #   - fixed epsg and empty_input
    other_suffixes = list(suffixes)
    other_suffixes.remove(".gpkg")
    for suffix in other_suffixes:
        for geoops_module in geoops_modules:
            for testfile in testfiles:
                where_post = ""
                keep_empty_geoms = False
                gridsize = 0.01 if testfile == "polygon-parcel" else GRIDSIZE_DEFAULT
                if testfile == "polygon-parcel":
                    where_post = WHERE_AREA_GT_400
                else:
                    keep_empty_geoms = True
                dimensions = None
                result.append(
                    (
                        suffix,
                        31370,
                        geoops_module,
                        testfile,
                        False,
                        gridsize,
                        keep_empty_geoms,
                        where_post,
                        dimensions,
                    )
                )

    # Test empty_input=True on
    #   - all combinations of fileops_modules and SUFFIXES
    #   - fixed epsg, testfile and empty_input
    for geoops_module in geoops_modules:
        for suffix in suffixes:
            gridsize = 0.01 if suffix == ".gpkg" else GRIDSIZE_DEFAULT
            keep_empty_geoms = False
            where_post = None
            dimensions = None
            result.append(
                (
                    suffix,
                    31370,
                    geoops_module,
                    "polygon-parcel",
                    True,
                    gridsize,
                    keep_empty_geoms,
                    where_post,
                    dimensions,
                )
            )

    return result


@pytest.mark.parametrize(
    "suffix, epsg, geoops_module, testfile, empty_input, gridsize, keep_empty_geoms, "
    "where_post, dimensions",
    basic_combinations_to_test(),
)
def test_buffer(
    tmp_path,
    suffix,
    epsg,
    geoops_module,
    testfile,
    empty_input,
    gridsize,
    keep_empty_geoms,
    where_post,
    dimensions,
):
    """Buffer basics are available both in the gpd and sql implementations."""
    # Prepare test data
    input_path = test_helper.get_testfile(
        testfile, suffix=suffix, epsg=epsg, empty=empty_input, dimensions=dimensions
    )

    # Now run test
    output_path = tmp_path / f"{input_path.stem}-{geoops_module}{suffix}"
    set_geoops_module(geoops_module)
    input_layerinfo = fileops.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)
    assert input_layerinfo.crs is not None
    distance = 1
    if input_layerinfo.crs.is_projected is False:
        # 1 degree = 111 km or 111000 m
        distance /= 111000

    # Prepare expected result
    expected_gdf = fileops.read_file(input_path)
    expected_gdf.geometry = expected_gdf.geometry.buffer(distance, resolution=5)
    # Default value for keep_empty_geoms is False
    keep_empty_geoms_prepped = False if keep_empty_geoms is None else keep_empty_geoms
    expected_gdf = test_helper.prepare_expected_result(
        expected_gdf,
        gridsize=gridsize,
        keep_empty_geoms=keep_empty_geoms_prepped,
        where_post=where_post,
    )

    # Test positive buffer
    kwargs = {}
    if keep_empty_geoms is not None:
        kwargs["keep_empty_geoms"] = keep_empty_geoms
    geoops.buffer(
        input_path=input_path,
        output_path=output_path,
        distance=distance,
        gridsize=gridsize,
        keep_empty_geoms=keep_empty_geoms,
        where_post=where_post,
        nb_parallel=2,
        batchsize=batchsize,
    )

    # Now check if the output file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert fileops.has_spatial_index(output_path) is exp_spatial_index
    output_layerinfo = fileops.get_layerinfo(output_path)
    assert len(output_layerinfo.columns) == len(input_layerinfo.columns)

    if empty_input:
        return

    # More detailed check
    output_gdf = fileops.read_file(output_path)
    assert output_gdf["geometry"][0] is not None
    check_less_precise = True if input_layerinfo.crs.is_projected is False else False
    assert_geodataframe_equal(
        output_gdf,
        expected_gdf,
        normalize=True,
        promote_to_multi=True,
        check_less_precise=check_less_precise,
        sort_values=True,
    )


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
@pytest.mark.parametrize("testfile", ["polygon-parcel"])
@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
def test_buffer_columns_fid(tmp_path, suffix, geoops_module, testfile):
    """Buffer basics are available both in the gpd and sql implementations."""
    # Prepare test data
    input_path = test_helper.get_testfile(testfile, suffix=suffix)

    # Now run test
    output_path = tmp_path / f"{input_path.stem}-{geoops_module}{suffix}"
    set_geoops_module(geoops_module)
    input_layerinfo = fileops.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Test positive buffer
    geoops.buffer(
        input_path=input_path,
        output_path=output_path,
        distance=1,
        columns=["LblHfdTlt", "fid"],
        explodecollections=True,
        keep_empty_geoms=False,
        nb_parallel=2,
        batchsize=batchsize,
    )

    # Read input file and extract some info
    input_gdf = fileops.read_file(input_path, fid_as_index=True)
    if _geofileinfo.get_geofileinfo(input_path).is_fid_zerobased:
        assert input_gdf.index[0] == 0
    else:
        assert input_gdf.index[0] == 1
    input_multi_gdf = input_gdf[
        input_gdf.geometry.buffer(0).geom_type == "MultiPolygon"
    ]
    assert len(input_multi_gdf) == 2
    multi_fid = input_multi_gdf.index[0]

    # Now check if the output file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert fileops.has_spatial_index(output_path) is exp_spatial_index
    output_layerinfo = fileops.get_layerinfo(output_path)
    output_gdf = fileops.read_file(output_path)
    assert output_gdf["geometry"][0] is not None
    assert list(output_layerinfo.columns) == ["LblHfdTlt", "fid_1"]
    assert len(output_gdf[output_gdf.fid_1 == multi_fid]) == 2


@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
def test_buffer_force(tmp_path, geoops_module):
    input_path = test_helper.get_testfile("polygon-parcel")
    input_layerinfo = fileops.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)
    distance = 1
    set_geoops_module(geoops_module)

    # Run buffer
    output_path = tmp_path / f"{input_path.stem}-output{input_path.suffix}"
    assert output_path.exists() is False

    geoops.buffer(
        input_path=input_path,
        output_path=output_path,
        distance=distance,
        keep_empty_geoms=False,
        nb_parallel=2,
        batchsize=batchsize,
    )

    # Test buffer to existing output path
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert fileops.has_spatial_index(output_path) is exp_spatial_index
    mtime_orig = output_path.stat().st_mtime
    geoops.buffer(
        input_path=input_path,
        output_path=output_path,
        distance=distance,
        keep_empty_geoms=False,
        nb_parallel=2,
        batchsize=batchsize,
    )
    assert output_path.stat().st_mtime == mtime_orig

    # With force=True
    geoops.buffer(
        input_path=input_path,
        output_path=output_path,
        distance=distance,
        keep_empty_geoms=False,
        nb_parallel=2,
        batchsize=batchsize,
        force=True,
    )
    assert output_path.stat().st_mtime != mtime_orig


@pytest.mark.parametrize(
    "expected_error, input_path, output_path",
    [
        (
            "buffer: output_path must not equal input_path",
            test_helper.get_testfile("polygon-parcel"),
            test_helper.get_testfile("polygon-parcel"),
        ),
        (
            "buffer: input_path doesn't exist:",
            "not_existing_path",
            "output.gpkg",
        ),
    ],
)
@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
def test_buffer_invalid_params(
    tmp_path, input_path, output_path, expected_error, geoops_module
):
    """
    Invalid params for single layer operations.
    """
    # Internal functions are directly called, so need to be Path objects
    if isinstance(output_path, str):
        output_path = tmp_path / output_path
    if isinstance(input_path, str):
        input_path = tmp_path / input_path

    # Now run test
    set_geoops_module(geoops_module)
    with pytest.raises(ValueError, match=expected_error):
        geoops.buffer(input_path=input_path, output_path=output_path, distance=1)


@pytest.mark.parametrize(
    "suffix, epsg, geoops_module, testfile, empty_input, gridsize, keep_empty_geoms, "
    "where_post, dimensions",
    basic_combinations_to_test(epsgs=[31370]),
)
def test_buffer_negative(
    tmp_path,
    suffix,
    epsg,
    geoops_module,
    testfile,
    empty_input,
    gridsize,
    keep_empty_geoms,
    where_post,
    dimensions,
):
    """Buffer basics are available both in the gpd and sql implementations."""
    input_path = test_helper.get_testfile(testfile, suffix=suffix)

    # Now run test
    output_path = tmp_path / f"{input_path.stem}-{geoops_module}{suffix}"
    set_geoops_module(geoops_module)
    input_layerinfo = fileops.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)
    keep_empty_geoms_prepped = False if keep_empty_geoms is None else keep_empty_geoms

    # Test negative buffer
    distance = -10
    kwargs = {}
    if keep_empty_geoms is not None:
        kwargs["keep_empty_geoms"] = keep_empty_geoms
    output_path = output_path.parent / f"{output_path.stem}_m10m{output_path.suffix}"
    geoops.buffer(
        input_path=input_path,
        output_path=output_path,
        distance=distance,
        gridsize=gridsize,
        where_post=where_post,
        nb_parallel=2,
        batchsize=batchsize,
        **kwargs,
    )

    # Now check if the output file is correctly created
    assert output_path.exists()

    output_layerinfo = fileops.get_layerinfo(output_path)
    assert len(output_layerinfo.columns) == len(input_layerinfo.columns)

    if input_layerinfo.geometrytype in [
        GeometryType.MULTIPOINT,
        GeometryType.MULTILINESTRING,
    ]:
        # A Negative buffer of points or linestrings gives NULL geometries
        if keep_empty_geoms_prepped:
            # If no filtering on NULL geoms, all rows are still present
            assert output_layerinfo.featurecount == input_layerinfo.featurecount
            if suffix != ".shp":
                # The None geoms aren't properly detected as geometry in shp, so becomes
                # an extra attribute column...
                # assert len(output_layerinfo.columns) == len(input_layerinfo.columns)
                pass
        else:
            # Everything is filtered away...
            assert output_layerinfo.featurecount == 0

    else:
        # A Negative buffer of polygons gives a result for large polygons.
        # 7 polygons disappear because of the negative buffer
        assert output_layerinfo.geometrytype == GeometryType.MULTIPOLYGON

        # Read result for some more detailed checks
        output_gdf = fileops.read_file(output_path)

        # Prepare expected result
        expected_gdf = fileops.read_file(input_path)
        expected_gdf.geometry = expected_gdf.geometry.buffer(distance, resolution=5)
        expected_gdf = test_helper.prepare_expected_result(
            expected_gdf,
            gridsize=gridsize,
            keep_empty_geoms=keep_empty_geoms_prepped,
            where_post=where_post,
        )
        assert_geodataframe_equal(output_gdf, expected_gdf, sort_values=True)


@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
def test_buffer_negative_explode(tmp_path, geoops_module):
    """Buffer basics are available both in the gpd and sql implementations."""
    input_path = test_helper.get_testfile("polygon-parcel")

    # Now run test
    output_path = tmp_path / f"{input_path.stem}-output{input_path.suffix}"
    set_geoops_module(geoops_module)
    input_layerinfo = fileops.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Test negative buffer with explodecollections
    output_path = (
        output_path.parent / f"{output_path.stem}_m10m_explode{output_path.suffix}"
    )
    distance = -10
    keep_empty_geoms = False
    geoops.buffer(
        input_path=input_path,
        output_path=output_path,
        distance=distance,
        explodecollections=True,
        keep_empty_geoms=keep_empty_geoms,
        nb_parallel=2,
        batchsize=batchsize,
    )

    # Now check if the output file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert fileops.has_spatial_index(output_path) is exp_spatial_index
    layerinfo_output = fileops.get_layerinfo(output_path)
    assert len(layerinfo_output.columns) == len(input_layerinfo.columns)
    assert layerinfo_output.geometrytype == GeometryType.POLYGON

    output_gdf = fileops.read_file(output_path)

    expected_gdf = fileops.read_file(input_path)
    expected_gdf.geometry = expected_gdf.geometry.buffer(distance, resolution=5)
    expected_gdf = test_helper.prepare_expected_result(
        expected_gdf, explodecollections=True, keep_empty_geoms=keep_empty_geoms
    )
    assert_geodataframe_equal(
        output_gdf,
        expected_gdf,
        promote_to_multi=True,
        sort_values=True,
        check_crs=False,
    )


@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
@pytest.mark.parametrize(
    "keep_empty_geoms, where_post", [(False, None), (False, WHERE_AREA_GT_400)]
)
@pytest.mark.parametrize("explodecollections", [True, False])
def test_buffer_negative_where_explode(
    tmp_path, suffix, geoops_module, keep_empty_geoms, where_post, explodecollections
):
    """Buffer basics are available both in the gpd and sql implementations."""
    # Prepare test data/environment
    input_path = test_helper.get_testfile("polygon-parcel", suffix=suffix)
    set_geoops_module(geoops_module)
    output_path = tmp_path / f"{input_path.stem}-{geoops_module}{suffix}"
    input_layerinfo = fileops.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)
    distance = -10

    expected_gdf = fileops.read_file(input_path)
    expected_gdf.geometry = expected_gdf.geometry.buffer(distance, resolution=5)
    expected_gdf = test_helper.prepare_expected_result(
        expected_gdf,
        keep_empty_geoms=keep_empty_geoms,
        where_post=where_post,
        explodecollections=explodecollections,
    )

    # Run test
    geoops.buffer(
        input_path=input_path,
        output_path=output_path,
        distance=distance,
        explodecollections=explodecollections,
        keep_empty_geoms=keep_empty_geoms,
        where_post=where_post,
        nb_parallel=2,
        batchsize=batchsize,
    )

    # Check result
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert fileops.has_spatial_index(output_path) is exp_spatial_index
    output_layerinfo = fileops.get_layerinfo(output_path)
    assert len(output_layerinfo.columns) == len(input_layerinfo.columns)

    output_gdf = fileops.read_file(output_path)
    assert output_gdf["geometry"][0] is not None

    assert_geodataframe_equal(
        output_gdf,
        expected_gdf,
        promote_to_multi=True,
        sort_values=True,
    )


@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
def test_buffer_preserve_fid_gpkg(tmp_path, geoops_module):
    """
    Buffer test to check if fid is properly preserved.

    Only relevant for e.g. Geopackage as it saves the fid as a value. For e.g.
    shapefiles the fid is not retained, but always a sequence from 0 in the file.
    """
    # Prepare test data: remove 2 fid's from file to check if the fid's are preserved
    input_full_path = test_helper.get_testfile("polygon-parcel")
    input_path = test_helper.get_testfile("polygon-parcel", dst_dir=tmp_path)
    input_layer = fileops.get_only_layer(input_path)
    sql_stmt = f'DELETE FROM "{input_layer}" WHERE fid IN (5, 6)'
    fileops.execute_sql(input_path, sql_stmt=sql_stmt)

    # Now run test
    output_path = tmp_path / f"{input_path.stem}-{geoops_module}.gpkg"
    set_geoops_module(geoops_module)
    input_layerinfo = fileops.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Test positive buffer
    geoops.buffer(
        input_path=input_path,
        output_path=output_path,
        distance=1,
        keep_empty_geoms=False,
        nb_parallel=2,
        batchsize=batchsize,
    )

    # Prepare expected result to compare with
    expected_gdf = fileops.read_file(input_full_path, fid_as_index=True)
    expected_gdf = expected_gdf[~expected_gdf.index.isin([5, 6])]
    expected_gdf = expected_gdf[~expected_gdf.geometry.is_empty]

    # Check if the output file with the expected result
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert fileops.has_spatial_index(output_path) is exp_spatial_index
    output_gdf = fileops.read_file(output_path, fid_as_index=True)
    assert len(output_gdf) == len(expected_gdf)
    assert output_gdf["geometry"].iloc[0] is not None
    assert (
        output_gdf.index.sort_values().tolist()
        == expected_gdf.index.sort_values().tolist()
    )


@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
def test_buffer_shp_to_gpkg(
    tmp_path,
    geoops_module,
):
    """
    Buffer from shapefile to gpkg.

    Test added because this gave a unique constraint on fid's, because for each partial
    file the fid started again with 0.
    """
    # Prepare test data
    input_path = test_helper.get_testfile("polygon-parcel", suffix=".shp")

    # Now run test
    output_path = tmp_path / f"{input_path.stem}-{geoops_module}.gpkg"
    set_geoops_module(geoops_module)
    input_layerinfo = fileops.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)
    distance = 1
    keep_empty_geoms = False

    # Prepare expected result
    expected_gdf = fileops.read_file(input_path)
    expected_gdf.geometry = expected_gdf.geometry.buffer(distance, resolution=5)
    expected_gdf = test_helper.prepare_expected_result(
        expected_gdf, keep_empty_geoms=keep_empty_geoms
    )

    # Test positive buffer
    geoops.buffer(
        input_path=input_path,
        output_path=output_path,
        distance=distance,
        nb_parallel=2,
        batchsize=batchsize,
        keep_empty_geoms=keep_empty_geoms,
    )

    # Now check if the output file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert fileops.has_spatial_index(output_path) is exp_spatial_index
    output_layerinfo = fileops.get_layerinfo(output_path)
    assert len(output_layerinfo.columns) == len(input_layerinfo.columns)

    # More detailed check
    output_gdf = fileops.read_file(output_path)
    assert output_gdf["geometry"][0] is not None
    assert_geodataframe_equal(
        output_gdf,
        expected_gdf,
        promote_to_multi=True,
        sort_values=True,
    )


@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
@pytest.mark.parametrize(
    "empty_input, gridsize, keep_empty_geoms, where_post",
    [(True, 0.0, True, None), (False, 0.01, None, WHERE_AREA_GT_400)],
)
def test_convexhull(
    tmp_path, geoops_module, suffix, empty_input, gridsize, keep_empty_geoms, where_post
):
    # Prepare test data
    logging.basicConfig(level=logging.DEBUG)
    input_path = test_helper.get_testfile(
        "polygon-parcel", suffix=suffix, empty=empty_input
    )
    set_geoops_module(geoops_module)
    input_layerinfo = fileops.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)

    # Also check if columns parameter works (case insensitive)
    columns = ["OIDN", "uidn", "HFDTLT", "lblhfdtlt", "GEWASGROEP", "lengte", "OPPERVL"]

    # Prepare expected result
    expected_gdf = fileops.read_file(input_path)
    expected_gdf.geometry = expected_gdf.geometry.convex_hull
    keep_empty_geoms_prepped = False if keep_empty_geoms is None else keep_empty_geoms
    expected_gdf = test_helper.prepare_expected_result(
        expected_gdf,
        gridsize=gridsize,
        keep_empty_geoms=keep_empty_geoms_prepped,
        where_post=where_post,
        columns=columns,
    )

    # Run test
    kwargs = {}
    if keep_empty_geoms is not None:
        kwargs["keep_empty_geoms"] = keep_empty_geoms
    output_path = tmp_path / f"{input_path.stem}-output{suffix}"
    geoops.convexhull(
        input_path=input_path,
        columns=columns,
        output_path=output_path,
        gridsize=gridsize,
        where_post=where_post,
        nb_parallel=2,
        batchsize=batchsize,
        **kwargs,
    )

    # Now check if the output file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert fileops.has_spatial_index(output_path) is exp_spatial_index
    layerinfo_output = fileops.get_layerinfo(output_path)
    assert "OIDN" in layerinfo_output.columns
    assert "uidn" in layerinfo_output.columns
    assert len(layerinfo_output.columns) == len(columns)
    assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON

    # If input was empty, we are already OK
    if empty_input:
        return

    # More detailed check
    output_gdf = fileops.read_file(output_path)
    assert output_gdf["geometry"][0] is not None
    assert_geodataframe_equal(output_gdf, expected_gdf, sort_values=True)


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
@pytest.mark.parametrize("input_empty", [True, False])
@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
def test_makevalid(tmp_path, suffix, input_empty, geoops_module):
    # Prepare test data
    input_path = test_helper.get_testfile(
        "polygon-invalid", suffix=suffix, empty=input_empty
    )
    set_geoops_module(geoops_module)

    # If the input file is not empty, it should have invalid geoms
    if not input_empty:
        input_isvalid_path = tmp_path / f"{input_path.stem}_is-valid{suffix}"
        isvalid = _geoops_sql.isvalid(
            input_path=input_path, output_path=input_isvalid_path
        )
        assert isvalid is False, "Input file should contain invalid features"

    # Make sure the input file is not valid
    if not input_empty:
        output_isvalid_path = (
            tmp_path / f"{input_path.stem}_is-valid{input_path.suffix}"
        )
        isvalid = _geoops_sql.isvalid(
            input_path=input_path, output_path=output_isvalid_path
        )
        assert isvalid is False, "Input file should contain invalid features"

    # Add some extra kwargs
    kwargs = {}
    if geoops_module == "geofileops.geoops":
        kwargs["validate_attribute_data"] = True

    # Do operation
    output_path = tmp_path / f"{input_path.stem}-output{suffix}"
    geoops.makevalid(
        input_path=input_path,
        output_path=output_path,
        nb_parallel=2,
        force_output_geometrytype=fileops.GeometryType.MULTIPOLYGON,
        **kwargs,
    )

    # Now check if the output file is correctly created
    assert output_path.exists()
    layerinfo_orig = fileops.get_layerinfo(input_path)
    layerinfo_output = fileops.get_layerinfo(output_path)
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)
    assert layerinfo_output.geometrytype in [
        GeometryType.POLYGON,
        GeometryType.MULTIPOLYGON,
    ]

    if not input_empty:
        assert layerinfo_orig.featurecount == layerinfo_output.featurecount

    # Check if the result file is valid
    output_new_isvalid_path = (
        tmp_path / f"{output_path.stem}_new_is-valid{output_path.suffix}"
    )
    isvalid = _geoops_sql.isvalid(
        input_path=output_path, output_path=output_new_isvalid_path
    )
    assert isvalid is True, "Output file shouldn't contain invalid features"

    # Run makevalid with existing output file and force=False (=default)
    geoops.makevalid(input_path=input_path, output_path=output_path)


@pytest.mark.parametrize(
    "descr, geometry, expected_geometry",
    [
        ("sliver", Polygon([(0, 0), (5, 0), (10, 0), (15, 0)]), Polygon()),
        (
            "poly + line",
            MultiPolygon(
                [
                    Polygon([(0, 5), (5, 5), (5, 10), (0, 10), (0, 5)]),
                    Polygon([(0, 0), (5, 0), (10, 0), (15, 0)]),
                ]
            ),
            Polygon([(0, 5), (5, 5), (5, 10), (0, 10), (0, 5)]),
        ),
    ],
)
@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
def test_makevalid_collapsing_part(
    tmp_path, descr: str, geometry, geoops_module, expected_geometry
):
    # Prepare test data
    set_geoops_module(geoops_module)
    input_gdf = gpd.GeoDataFrame({"descr": [descr]}, geometry=[geometry], crs=31370)
    input_path = tmp_path / "test.gpkg"
    fileops.to_file(input_gdf, input_path)

    # Now we are ready to test
    result_path = tmp_path / "test_makevalid_collapsing_part.gpkg"
    geoops.makevalid(
        input_path=input_path,
        output_path=result_path,
        keep_empty_geoms=False,
        force=True,
    )
    result_gdf = fileops.read_file(result_path)

    # Compare with expected result
    expected_gdf = gpd.GeoDataFrame(
        {"descr": [descr]}, geometry=[expected_geometry], crs=31370
    )
    expected_gdf = expected_gdf[~expected_gdf.geometry.is_empty]
    if len(expected_gdf) == 0:
        assert len(result_gdf) == 0
    else:
        assert_geodataframe_equal(result_gdf, expected_gdf)


@pytest.mark.parametrize("suffix", SUFFIXES_GEOOPS)
@pytest.mark.parametrize("explodecollections", [True, False])
@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
def test_makevalid_exploded_input(tmp_path, suffix, geoops_module, explodecollections):
    """
    Test that even if single polygon input, output is multipolygon.

    Unless explodecollections of the output is True. Is necessary because makevalid can
    create multipolygons, and then the output becomes GEOMETRY.
    """
    # Prepare test data
    input_path = test_helper.get_testfile(
        "polygon-invalid", suffix=suffix, explodecollections=True
    )
    set_geoops_module(geoops_module)

    # Do operation
    output_path = tmp_path / f"{input_path.stem}-output{suffix}"
    geoops.makevalid(
        input_path=input_path,
        output_path=output_path,
        explodecollections=explodecollections,
        nb_parallel=2,
    )

    # Now check if the output file is correctly created
    assert output_path.exists()
    layerinfo_orig = fileops.get_layerinfo(input_path)
    layerinfo_output = fileops.get_layerinfo(output_path)
    assert len(layerinfo_orig.columns) == len(layerinfo_output.columns)
    if explodecollections and suffix != ".shp":
        assert layerinfo_output.geometrytype == GeometryType.POLYGON
    else:
        assert layerinfo_output.geometrytype == GeometryType.MULTIPOLYGON


@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
@pytest.mark.parametrize("gridsize", [0.0, 0.01])
@pytest.mark.parametrize("keep_empty_geoms", [None, True])
def test_makevalid_gridsize(tmp_path, geoops_module, gridsize, keep_empty_geoms):
    """
    Apply gridsize on the default test file to make it removes sliver polygon.
    """
    # Prepare test data
    input_path = test_helper.get_testfile("polygon-parcel")
    input_info = fileops.get_layerinfo(input_path)

    expected_featurecount = input_info.featurecount
    # If NULL/EMPTY geoms should not be kept, the expected featurecount is lower.
    # (keep_empty_geoms=False is the default)
    if not keep_empty_geoms or keep_empty_geoms is None:
        expected_featurecount -= 1
        # With gridsize specified, a sliver polygon is removed as well
        if gridsize > 0.0:
            # If sql based and spatialite < 5.1, the sliver isn't cleaned up...
            if not (
                not SPATIALITE_GTE_51 and geoops_module == "geofileops.util._geoops_sql"
            ):
                expected_featurecount -= 1

    set_geoops_module(geoops_module)

    # Do operation
    kwargs = {}
    if keep_empty_geoms is not None:
        kwargs["keep_empty_geoms"] = keep_empty_geoms
    output_path = tmp_path / f"{input_path.stem}-output-{gridsize}.gpkg"
    geoops.makevalid(
        input_path=input_path,
        output_path=output_path,
        gridsize=gridsize,
        nb_parallel=2,
        **kwargs,
    )

    output_info = fileops.get_layerinfo(output_path)
    assert output_info.featurecount == expected_featurecount


@pytest.mark.parametrize(
    "descr, geometry, expected_geometry",
    [
        ("sliver", Polygon([(0, 0), (10, 0), (10, 0.5), (0, 0)]), Polygon()),
        (
            "poly + sliver",
            MultiPolygon(
                [
                    Polygon([(0, 5), (5, 5), (5, 10), (0, 10), (0, 5)]),
                    Polygon([(0, 0), (10, 0), (10, 0.5), (0, 0)]),
                ]
            ),
            Polygon([(0, 5), (5, 5), (5, 10), (0, 10), (0, 5)]),
        ),
    ],
)
@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
def test_makevalid_gridsize_extra(
    tmp_path, descr: str, geometry, geoops_module, expected_geometry
):
    # Prepare test data
    set_geoops_module(geoops_module)
    input_gdf = gpd.GeoDataFrame({"descr": [descr]}, geometry=[geometry], crs=31370)
    input_path = tmp_path / "test.gpkg"
    fileops.to_file(input_gdf, input_path)
    gridsize = 1

    # Now we are ready to test
    result_path = tmp_path / "test_makevalid.gpkg"
    geoops.makevalid(
        input_path=input_path,
        output_path=result_path,
        gridsize=gridsize,
        keep_empty_geoms=False,
        force=True,
    )
    result_gdf = fileops.read_file(result_path)

    # Compare with expected result
    expected_gdf = gpd.GeoDataFrame(
        {"descr": [descr]}, geometry=[expected_geometry], crs=31370
    )
    expected_gdf = expected_gdf[~expected_gdf.geometry.is_empty]
    if len(expected_gdf) == 0:
        assert len(result_gdf) == 0
    else:
        assert_geodataframe_equal(result_gdf, expected_gdf)


@pytest.mark.parametrize("geoops_module", GEOOPS_MODULES)
def test_makevalid_gridsize_topoerror(tmp_path, geoops_module):
    """
    Test on a specific valid polygon that gives a topologyerror when gridsize is set.
    """
    # The only currently known test case only works with geos 3.12. sql implementation
    # doesn't handle this case properly.
    if (
        shapely.geos_version != (3, 12, 0)
        or geoops_module == "geofileops.util._geoops_sql"
    ):
        pytest.skip()

    # Prepare test data
    poly_gridsize_error = shapely.from_wkt(
        "Polygon ((26352.5 175096.6, 26352.6 175096.6, 26352.6 175096.7, "
        "26352.5 175096.7, 26352.5 175096.6),(26352.528000000002 175096.676, "
        "26352.528369565214 175096.67489130437, 26352.528140495866 175096.67619834712, "
        "26352.52785714286 175096.67714285714, 26352.53 175096.66, "
        "26352.528000000002 175096.676))"
    )
    poly_ok = shapely.from_wkt(
        "Polygon ((26352.5 175096.7, 26352.66895 175096.76895, 26352.6 175096.6, "
        "26352.5 175096.6, 26352.5 175096.7))"
    )
    test_data = {
        "descr": ["gridsize_error", "ok"],
        "geometry": [poly_gridsize_error, poly_ok],
    }
    test_gdf = gpd.GeoDataFrame(test_data, crs=31370)
    input_path = tmp_path / "input.gpkg"
    fileops.to_file(test_gdf, input_path)

    gridsize = 0.01

    # Prepare expected result
    expected_data = {
        "descr": ["gridsize_error", "ok"],
        "geometry": [
            poly_gridsize_error,
            shapely.set_precision(poly_ok, grid_size=gridsize),
        ],
    }
    assert test_data["geometry"][1] != expected_data["geometry"][1]
    expected_gdf = gpd.GeoDataFrame(expected_data, crs=31370)

    # Now run test
    set_geoops_module(geoops_module)
    output_path = tmp_path / "output.gpkg"
    geoops.makevalid(input_path=input_path, output_path=output_path, gridsize=gridsize)

    # Check result
    output_gdf = fileops.read_file(output_path)
    assert_geodataframe_equal(
        output_gdf, expected_gdf, promote_to_multi=True, normalize=True
    )


def test_makevalid_invalidparams():
    # Only test on geoops, as the invalid params only exist there.
    set_geoops_module("geofileops.geoops")
    expected_error = (
        "the precision parameter is deprecated and cannot be combined with gridsize"
    )
    with pytest.raises(ValueError, match=expected_error):
        geoops.makevalid(
            input_path="abc",
            output_path="def",
            gridsize=1,
            precision=1,
        )


@pytest.mark.parametrize(
    "suffix, epsg, geoops_module, testfile, empty_input, gridsize, keep_empty_geoms, "
    "where_post, dimensions",
    basic_combinations_to_test(testfiles=["polygon-parcel", "linestring-row-trees"]),
)
def test_simplify(
    tmp_path,
    suffix,
    epsg,
    geoops_module,
    testfile,
    empty_input,
    gridsize,
    keep_empty_geoms,
    where_post,
    dimensions,
):
    # Prepare test data
    tmp_dir = tmp_path / f"{geoops_module}_{epsg}"
    tmp_dir.mkdir(parents=True, exist_ok=True)
    input_path = test_helper.get_testfile(
        testfile, dst_dir=tmp_dir, suffix=suffix, epsg=epsg, empty=empty_input
    )
    output_path = tmp_dir / f"{input_path.stem}-output{suffix}"
    set_geoops_module(geoops_module)
    input_layerinfo = fileops.get_layerinfo(input_path)
    batchsize = math.ceil(input_layerinfo.featurecount / 2)
    assert input_layerinfo.crs is not None
    if input_layerinfo.crs.is_projected:
        tolerance = 5
    else:
        # 1 degree = 111 km or 111000 m
        tolerance = 5 / 111000
    keep_empty_geoms_prepped = False if keep_empty_geoms is None else keep_empty_geoms

    # Prepare expected result
    expected_gdf = fileops.read_file(input_path)
    expected_gdf.geometry = expected_gdf.geometry.simplify(tolerance)
    expected_gdf = test_helper.prepare_expected_result(
        expected_gdf,
        gridsize=gridsize,
        keep_empty_geoms=keep_empty_geoms_prepped,
        where_post=where_post,
    )

    # Test default algorithm (rdp)
    kwargs = {}
    if keep_empty_geoms is not None:
        kwargs["keep_empty_geoms"] = keep_empty_geoms
    output_path = io_util.with_stem(input_path, output_path)
    geoops.simplify(
        input_path=input_path,
        output_path=output_path,
        tolerance=tolerance,
        gridsize=gridsize,
        where_post=where_post,
        nb_parallel=2,
        batchsize=batchsize,
        **kwargs,
    )

    # Now check if the tmp file is correctly created
    assert output_path.exists()
    exp_spatial_index = GeofileInfo(output_path).default_spatial_index
    assert fileops.has_spatial_index(output_path) is exp_spatial_index
    output_layerinfo = fileops.get_layerinfo(output_path)
    assert len(output_layerinfo.columns) == len(input_layerinfo.columns)

    # If input was empty, we are already OK
    if empty_input:
        return

    # More detailed checks
    output_gdf = fileops.read_file(output_path)
    assert_geodataframe_equal(
        output_gdf, expected_gdf, sort_values=True, normalize=True
    )


@pytest.mark.parametrize(
    "algorithm",
    [
        "lang",
        "lang+",
        "rdp",
        "vw",
        geoops.SimplifyAlgorithm.LANG,
        geoops.SimplifyAlgorithm.LANGP,
        geoops.SimplifyAlgorithm.RAMER_DOUGLAS_PEUCKER,
        geoops.SimplifyAlgorithm.VISVALINGAM_WHYATT,
    ],
)
def test_simplify_algorithms(tmp_path, algorithm):
    """
    Rude check on supported algorithms.
    """
    input_path = test_helper.get_testfile("polygon-parcel")
    output_path = tmp_path / f"{input_path.stem}_output.gpkg"

    # Test specifically with geoops
    set_geoops_module("geofileops.geoops")
    geoops.simplify(
        input_path=input_path,
        output_path=output_path,
        tolerance=1,
        algorithm=algorithm,
        nb_parallel=2,
    )

    assert output_path.exists()



"""
Helper functions for all tests.
"""

import os
import tempfile
from pathlib import Path
from typing import Optional, Union

import geopandas as gpd
import geopandas.testing as gpd_testing
import shapely
import shapely.geometry as sh_geom

import geofileops as gfo
from geofileops.util import _geofileinfo, _geoseries_util, geodataframe_util

_data_dir = Path(__file__).parent.resolve() / "data"
EPSGS = [31370, 4326]
GRIDSIZE_DEFAULT = 0.0
SUFFIXES_FILEOPS = [".gpkg", ".shp", ".csv"]
SUFFIXES_GEOOPS = [".gpkg", ".shp"]
TESTFILES = ["polygon-parcel", "linestring-row-trees", "point"]
WHERE_AREA_GT_400 = "ST_Area({geometrycolumn}) > 400"
WHERE_AREA_GT_5000 = "ST_Area({geometrycolumn}) > 5000"
WHERE_LENGTH_GT_1000 = "ST_Length({geometrycolumn}) > 1000"
WHERE_LENGTH_GT_200000 = "ST_Length({geometrycolumn}) > 200000"

RUNS_LOCAL = True
if "GITHUB_ACTIONS" in os.environ:
    RUNS_LOCAL = False


def prepare_expected_result(
    gdf: gpd.GeoDataFrame,
    keep_empty_geoms: bool,
    gridsize: float = 0.0,
    where_post: Optional[str] = None,
    explodecollections=False,
    columns: Optional[list[str]] = None,
) -> gpd.GeoDataFrame:
    """Prepare expected data"""
    if keep_empty_geoms is None:
        raise ValueError("keep_empty_geoms cannot be None")

    exp_gdf = gdf.copy()

    if gridsize != 0.0:
        exp_gdf.geometry = shapely.set_precision(exp_gdf.geometry, grid_size=gridsize)
    if explodecollections:
        exp_gdf = exp_gdf.explode(ignore_index=True)

    # Check what filtering is needed
    filter_area_gt = None
    if where_post is None or where_post == "":
        pass
    elif where_post == "ST_Area({geometrycolumn}) > 400":
        filter_area_gt = 400
    elif where_post == "ST_Area({geometrycolumn}) > 5000":
        filter_area_gt = 5000
    else:
        raise ValueError(f"unsupported where_post parameter in test: {where_post}")

    # Apply filtering
    # If keep_empty_geoms is None, treat it as False as this is the default value
    if not keep_empty_geoms:
        exp_gdf = exp_gdf[~exp_gdf.geometry.isna()]
        exp_gdf = exp_gdf[~exp_gdf.geometry.is_empty]
    if filter_area_gt is not None:
        exp_gdf = exp_gdf[exp_gdf.geometry.area > filter_area_gt]

    if columns is not None:
        column_mapper = {}
        columns_to_drop = []
        columns_dict = {column.upper(): column for column in columns}
        for column in exp_gdf.columns:
            if column.upper() in columns_dict:
                column_mapper[column] = columns_dict[column.upper()]
            elif column != "geometry":
                columns_to_drop.append(column)
        exp_gdf = exp_gdf.rename(columns=column_mapper)
        if len(columns_to_drop) > 0:
            exp_gdf = exp_gdf.drop(columns=columns_to_drop)

    assert isinstance(exp_gdf, gpd.GeoDataFrame)
    return exp_gdf


def prepare_test_file(
    input_path: Path,
    output_dir: Path,
    suffix: str,
    crs_epsg: Optional[int] = None,
    use_cachedir: bool = False,
) -> Path:
    # Tmp dir
    if use_cachedir is True:
        tmp_cache_dir = Path(tempfile.gettempdir()) / "geofileops_test_data"
        tmp_cache_dir.mkdir(parents=True, exist_ok=True)
    else:
        tmp_cache_dir = output_dir

    # If crs_epsg specified and test input file in wrong crs_epsg, reproject
    input_prepared_path = input_path
    if crs_epsg is not None:
        input_prepared_path = tmp_cache_dir / f"{input_path.stem}_{crs_epsg}{suffix}"
        if input_prepared_path.exists() is False:
            input_layerinfo = gfo.get_layerinfo(input_path)
            assert input_layerinfo.crs is not None
            if input_layerinfo.crs.to_epsg() == crs_epsg:
                if input_path.suffix == suffix:
                    gfo.copy(input_path, input_prepared_path)
                else:
                    gfo.copy_layer(input_path, input_prepared_path)
            else:
                test_gdf = gfo.read_file(input_path)
                test_gdf = test_gdf.to_crs(crs_epsg)
                assert isinstance(test_gdf, gpd.GeoDataFrame)
                gfo.to_file(test_gdf, input_prepared_path)
    elif input_path.suffix != suffix:
        # No crs specified, but different suffix asked, so convert file
        input_prepared_path = tmp_cache_dir / f"{input_path.stem}{suffix}"
        if input_prepared_path.exists() is False:
            gfo.copy_layer(input_path, input_prepared_path)

    # Now copy the prepared file to the output dir
    output_path = output_dir / input_prepared_path.name
    if str(input_prepared_path) != str(output_path):
        gfo.copy(input_prepared_path, output_path)
    return output_path


def get_testfile(
    testfile: str,
    dst_dir: Optional[Path] = None,
    suffix: str = ".gpkg",
    epsg: int = 31370,
    empty: bool = False,
    dimensions: Optional[str] = None,
    explodecollections: bool = False,
) -> Path:
    # Prepare original filepath; but try first with .zip.gpkg file.
    testfile_path = _data_dir / f"{testfile}.gpkg"
    if not testfile_path.exists():
        raise ValueError(f"Invalid testfile type: {testfile}")

    # Prepare destination location
    if dst_dir is None:
        dst_dir = Path(tempfile.gettempdir()) / "geofileops_test_data"
    assert isinstance(dst_dir, Path)
    dst_dir.mkdir(parents=True, exist_ok=True)

    # Prepare file + return
    empty_str = "_empty" if empty else ""
    prepared_path = (
        dst_dir / f"{testfile_path.stem}_{epsg}_{dimensions}{empty_str}{suffix}"
    )
    if prepared_path.exists():
        return prepared_path
    layers = gfo.listlayers(testfile_path)
    dst_info = _geofileinfo.get_geofileinfo(prepared_path)
    if len(layers) > 1 and dst_info.is_singlelayer:
        raise ValueError(
            f"multilayer testfile ({testfile}) cannot be converted to single layer "
            f"geofiletype: {dst_info.driver}"
        )

    # Convert all layers found
    for src_layer in layers:
        # Single layer files have stem as layername
        dst_layer = prepared_path.stem if dst_info.is_singlelayer else src_layer

        gfo.copy_layer(
            testfile_path,
            prepared_path,
            src_layer=src_layer,
            dst_layer=dst_layer,
            dst_crs=epsg,
            reproject=True,
            append=True,
            preserve_fid=not explodecollections,
            dst_dimensions=dimensions,
            explodecollections=explodecollections,
        )

        if empty:
            # Remove all rows from destination layer.
            # GDAL only supports DELETE using SQLITE dialect, not with OGRSQL.
            gfo.execute_sql(
                prepared_path,
                sql_stmt=f'DELETE FROM "{dst_layer}"',
                sql_dialect="SQLITE",
            )
        elif dimensions is not None:
            if dimensions != "XYZ":
                raise ValueError(f"unimplemented dimensions: {dimensions}")

            prepared_info = gfo.get_layerinfo(
                prepared_path, layer=dst_layer, raise_on_nogeom=False
            )
            if prepared_info.geometrycolumn is not None:
                gfo.update_column(
                    prepared_path,
                    name=prepared_info.geometrycolumn,
                    expression=f"CastToXYZ({prepared_info.geometrycolumn}, 5.0)",
                    layer=dst_layer,
                )

    return prepared_path


class TestData:
    crs_epsg = 31370
    point = sh_geom.Point((0, 0))
    multipoint = sh_geom.MultiPoint([(0, 0), (10, 10), (20, 20)])
    linestring = sh_geom.LineString([(0, 0), (10, 10), (20, 20)])
    multilinestring = sh_geom.MultiLineString(
        [linestring.coords, [(100, 100), (110, 110), (120, 120)]]
    )
    polygon_with_island = sh_geom.Polygon(
        shell=[(0.01, 0), (0.01, 10), (1, 10), (10, 10), (10, 0), (0.01, 0)],
        holes=[[(2, 2), (2, 8), (8, 8), (8, 2), (2, 2)]],
    )
    polygon_no_islands = sh_geom.Polygon(
        shell=[(100.01, 100), (100.01, 110), (110, 110), (110, 100), (100.01, 100)]
    )
    polygon_with_island2 = sh_geom.Polygon(
        shell=[(20, 20), (20, 30), (21, 30), (30, 30), (30, 20), (20, 20)],
        holes=[[(22, 22), (22, 28), (28, 28), (28, 22), (22, 22)]],
    )
    multipolygon = sh_geom.MultiPolygon([polygon_no_islands, polygon_with_island2])
    geometrycollection = sh_geom.GeometryCollection(
        [
            point,
            multipoint,
            linestring,
            multilinestring,
            polygon_with_island,
            multipolygon,
        ]
    )
    polygon_small_island = sh_geom.Polygon(
        shell=[(40, 40), (40, 50), (41, 50), (50, 50), (50, 40), (40, 40)],
        holes=[[(42, 42), (42, 43), (43, 43), (43, 42), (42, 42)]],
    )


def create_tempdir(base_dirname: str, parent_dir: Optional[Path] = None) -> Path:
    # Parent
    if parent_dir is None:
        parent_dir = Path(tempfile.gettempdir())

    for i in range(1, 999999):
        try:
            tempdir = parent_dir / f"{base_dirname}_{i:06d}"
            tempdir.mkdir(parents=True)
            return Path(tempdir)
        except FileExistsError:
            continue

    raise Exception(
        "Wasn't able to create a temporary dir with basedir: "
        f"{parent_dir / base_dirname}"
    )


def assert_geodataframe_equal(
    left: gpd.GeoDataFrame,
    right: gpd.GeoDataFrame,
    check_dtype=True,
    check_index_type: Union[bool, str] = "equiv",
    check_column_type: Union[bool, str] = "equiv",
    check_frame_type=True,
    check_like=False,
    check_less_precise=False,
    check_geom_type=False,
    check_geom_empty_vs_None=True,
    check_crs=True,
    normalize=False,
    promote_to_multi=False,
    sort_columns=False,
    sort_values=False,
    output_dir: Optional[Path] = None,
):
    """
    Check that two GeoDataFrames are equal/

    Parameters
    ----------
    left, right : two GeoDataFrames
    check_dtype : bool, default True
        Whether to check the DataFrame dtype is identical.
    check_index_type, check_column_type : bool, default 'equiv'
        Check that index types are equal.
    check_frame_type : bool, default True
        Check that both are same type (*and* are GeoDataFrames). If False,
        will attempt to convert both into GeoDataFrame.
    check_like : bool, default False
        If true, ignore the order of rows & columns
    check_less_precise : bool, default False
        If True, use geom_almost_equals. if False, use geom_equals.
    check_geom_type : bool, default False
        If True, check that all the geom types are equal.
    check_geom_empty_vs_None : bool, default True
        If False, ignore differences between empty and None geometries.
    check_crs: bool, default True
        If `check_frame_type` is True, then also check that the
        crs matches.
    normalize: bool, default False
        If True, normalize the geometries before comparing equality.
        Typically useful with ``check_less_precise=True``, which uses
        ``geom_almost_equals`` and requires exact coordinate order.
    promote_to_multi: bool, default False
        If True, promotes to multi.
    sort_columns: bool, default False
        If True, sort the columns of the dataframe before compare.
    sort_values: bool, default False
        If True, sort the values of the geodataframe, including the geometry
        (as WKT).
    output_dir: Path, default None
        If not None, the left and right dataframes will be written to the
        directory as geojson files. If normalize, promote_to_multi and/or
        sort_values are True, the will be applied before writing.
    """
    if sort_columns:
        left = left[sorted(left.columns)]
        right = right[sorted(right.columns)]

    if not check_geom_empty_vs_None:
        # Set empty geoms to None for both inputs
        left = left.copy()
        left.loc[left.geometry.is_empty, ["geometry"]] = None
        right = right.copy()
        right.loc[right.geometry.is_empty, ["geometry"]] = None

    if promote_to_multi:
        left.geometry = _geoseries_util.harmonize_geometrytypes(
            left.geometry, force_multitype=True
        )
        right.geometry = _geoseries_util.harmonize_geometrytypes(
            right.geometry, force_multitype=True
        )

    if normalize:
        left.geometry = gpd.GeoSeries(
            shapely.normalize(left.geometry), index=left.index
        )
        right.geometry = gpd.GeoSeries(
            shapely.normalize(right.geometry),
            index=right.index,
        )

    if sort_values:
        left = geodataframe_util.sort_values(left).reset_index(drop=True)
        right = geodataframe_util.sort_values(right).reset_index(drop=True)

    if output_dir is not None:
        output_dir.mkdir(parents=True, exist_ok=True)
        output_path = output_dir / "left.geojson"
        gfo.to_file(left, output_path, create_spatial_index=None)
        output_path = output_dir / "right.geojson"
        gfo.to_file(right, output_path, create_spatial_index=None)

    gpd_testing.assert_geodataframe_equal(
        left=left,
        right=right,
        check_dtype=check_dtype,
        check_index_type=check_index_type,
        check_column_type=check_column_type,
        check_frame_type=check_frame_type,
        check_like=check_like,
        check_less_precise=check_less_precise,
        check_geom_type=check_geom_type,
        check_crs=check_crs,
        normalize=normalize,
    )



import pytest

import geofileops as gfo
from geofileops.util import _geoops_sql
from tests import test_helper


@pytest.mark.parametrize(
    "descr, nb_rows_input_layer, nb_parallel, batchsize, is_twolayer_operation, "
    "exp_nb_parallel, exp_nb_batches",
    [
        ("0 input rows, batchsize=1, singlelayer", 0, 2, 1, 0, 1, 1),
        ("0 input rows, nb_parallel=2, singlelayer", 0, 2, 0, 0, 1, 1),
        ("0 input rows, singlelayer", 0, -1, -1, 0, 1, 1),
        ("0 input rows, twolayer", 0, -1, -1, 1, 1, 1),
        ("1 input row, batchsize=1, singlelayer", 1, 2, 1, 0, 1, 1),
        ("1 input rows, nb_parallel=2, singlelayer", 1, 2, 0, 0, 1, 1),
        ("1 input rows, singlelayer", 1, -1, -1, 0, 1, 1),
        ("1 input rows, twolayer", 1, -1, -1, 1, 1, 1),
        ("1 input rows, singlelayer", 1, -1, -1, 0, 1, 1),
        ("2 input rows, nb_parallel=10, singlelayer", 2, 10, -1, 0, 2, 2),
        ("100 input row, batchsize=20, singlelayer", 100, -1, 20, 0, 5, 5),
        ("100 input rows, nb_parallel=2, singlelayer", 100, 2, 0, 0, 2, 2),
        ("100 input rows, nb_parallel=1, singlelayer", 100, 1, 0, 0, 1, 1),
        ("100 input rows, singlelayer", 100, -1, -1, 0, 1, 1),
        ("100 input rows, twolayer", 100, -1, -1, 1, 1, 1),
        ("1000 input row, batchsize=20, singlelayer", 1000, -1, 20, 0, 8, 56),
        ("1000 input rows, nb_parallel=2, singlelayer", 1000, 2, 0, 0, 2, 2),
        ("1000 input rows, nb_parallel=1, singlelayer", 1000, 1, 0, 0, 1, 1),
        ("1000 input rows, singlelayer", 1000, -1, -1, 0, 8, 8),
        ("1000 input rows, twolayer", 1000, -1, -1, 1, 8, 16),
    ],
)
def test_determine_nb_batches(
    descr: str,
    nb_rows_input_layer: int,
    nb_parallel: int,
    batchsize: int,
    is_twolayer_operation: bool,
    exp_nb_parallel: int,
    exp_nb_batches: int,
):
    res_nb_parallel, res_nb_batches = _geoops_sql._determine_nb_batches(
        nb_rows_input_layer=nb_rows_input_layer,
        nb_parallel=nb_parallel,
        batchsize=batchsize,
        is_twolayer_operation=is_twolayer_operation,
        cpu_count=8,
    )
    assert exp_nb_parallel == res_nb_parallel
    assert exp_nb_batches == res_nb_batches


@pytest.mark.parametrize(
    "input1_suffix, input2_suffix, output1_suffix, output2_suffix",
    [
        (".gpkg", None, ".gpkg", None),
        (".gpkg", ".gpkg", ".gpkg", ".gpkg"),
        (".gpkg", ".shp", ".gpkg", ".gpkg"),
        (".gpkg", ".sqlite", ".gpkg", ".gpkg"),
        (".shp", None, ".gpkg", None),
        (".shp", ".gpkg", ".gpkg", ".gpkg"),
        (".shp", ".sqlite", ".sqlite", ".sqlite"),
        (".sqlite", None, ".sqlite", None),
        (".sqlite", ".shp", ".sqlite", ".sqlite"),
        (".sqlite", ".sqlite", ".sqlite", ".sqlite"),
    ],
)
def test_prepare_processing_params_filetypes(
    tmp_path, input1_suffix, input2_suffix, output1_suffix, output2_suffix
):
    input1_path = test_helper.get_testfile("polygon-parcel", suffix=input1_suffix)
    input1_layer = gfo.get_only_layer(input1_path)
    input2_path = None
    input2_layer = None
    if input2_suffix is not None:
        input2_path = test_helper.get_testfile("polygon-parcel", suffix=input2_suffix)
        input2_layer = gfo.get_only_layer(input2_path)

    params = _geoops_sql._prepare_processing_params(
        input1_path=input1_path,
        input1_layer=input1_layer,
        input2_path=input2_path,
        input2_layer=input2_layer,
        tempdir=tmp_path,
        convert_to_spatialite_based=True,
        nb_parallel=2,
    )

    assert params is not None
    assert params.input1_path.suffix == output1_suffix
    assert params.input1_path.exists()
    assert params.input1_layer in gfo.listlayers(params.input1_path)

    # If the file format hasn't changed, the file should not be copied
    if input1_path.suffix == params.input1_path.suffix:
        assert input1_path == params.input1_path

    if input2_suffix is not None:
        assert params.input2_path.exists()
        assert params.input2_layer in gfo.listlayers(params.input2_path)
        assert params.input2_path.suffix == output2_suffix
        # If the file format hasn't changed, the file should not be copied
        if input2_path.suffix == params.input2_path.suffix:
            assert input2_path == params.input2_path
        # If both input files were copied, they should have been copied to seperate
        # files
        if params.input1_path.parent == tmp_path and input2_suffix is not None:
            assert params.input1_path != params.input2_path
    else:
        assert params.input2_path is None


@pytest.mark.parametrize(
    "desc, testfile, subdivide_coords, retval_None",
    [
        ("input not complex", "polygon-zone", 1000, True),
        ("input poly+complex", "polygon-zone", 1, False),
        ("input no poly", "linestring-watercourse", 1, True),
    ],
)
def test_subdivide_layer(desc, tmp_path, testfile, subdivide_coords, retval_None):
    path = test_helper.get_testfile(testfile)
    result = _geoops_sql._subdivide_layer(
        path=path,
        layer=None,
        output_dir=tmp_path,
        subdivide_coords=subdivide_coords,
        overlay_self=False,
    )

    if retval_None:
        assert result is None
    else:
        assert result is not None



"""
Tests for functionalities in _processing_util.
"""

import os

from geofileops.util import _processing_util


def test_processnice():
    # Test setting and getting some values for nice
    # Remark: the nice values tests are spcifically written to accomodate for
    # windows specificalities:
    #     - windows only supports 6 niceness classes. setprocessnice en
    #       getprocessnice maps niceness values to these classes.
    #     - when setting REALTIME priority (-20 niceness) apparently this
    #       results only to HIGH priority.
    nice_orig = _processing_util.getprocessnice()
    for niceness in [-15, -10, 0, 10, 19]:
        # Decreasing niceness (sometimes) isn't possible on linux, so skip
        if os.name != "nt":
            continue

        # Test!
        _processing_util.setprocessnice(niceness)
        nice = _processing_util.getprocessnice()
        assert nice == niceness

    # Test invalid values for nice value
    try:
        _processing_util.setprocessnice(20)
        exception_raised = False
    except ValueError:
        exception_raised = True
    assert exception_raised is True
    try:
        _processing_util.setprocessnice(-21)
        exception_raised = False
    except ValueError:
        exception_raised = True
    assert exception_raised is True

    # Reset niceness to original value before test
    _processing_util.setprocessnice(nice_orig)



import logging

import pytest

from geofileops.util import _geoops_gpd


@pytest.mark.parametrize(
    "descr, nb_rows_input_layer, nb_parallel, batchsize, cpu_count, bytes_usable, "
    "exp_nb_parallel, exp_nb_batches",
    [
        ("0 input rows, batchsize=1", 0, 2, 1, 8, 500_000_000, 1, 1),
        ("0 input rows, nb_parallel=2", 0, 2, 0, 8, 500_000_000, 1, 1),
        ("0 input rows", 0, -1, -1, 8, 500_000_000, 1, 1),
        ("1 input row, batchsize=1", 1, 2, 1, 8, 500_000_000, 1, 1),
        ("1 input rows, nb_parallel=2", 1, 2, 0, 8, 500_000_000, 1, 1),
        ("1 input rows", 1, -1, -1, 8, 500_000_000, 1, 1),
        ("2 input rows, nb_parallel=10", 2, 10, -1, 8, 500_000_000, 1, 1),
        ("100 input row, batchsize=20", 100, -1, 20, 8, 500_000_000, 5, 5),
        ("100 input rows", 100, -1, -1, 8, 500_000_000, 1, 1),
        ("1000 input row, batchsize=20", 1000, -1, 20, 8, 500_000_000, 8, 50),
        ("1000 input rows", 1000, -1, -1, 8, 500_000_000, 1, 1),
        ("2000 input rows", 2000, -1, -1, 8, 500_000_000, 2, 2),
        ("100_000 input rows", 100000, -1, -1, 8, 500_000_000, 8, 8),
        ("100_000 input rows, 250MB RAM", 1_000_000, -1, -1, 8, 250_000_000, 4, 28),
        ("100_000 input rows, 2CPU", 1_000_000, -1, -1, 2, 500_000_000, 2, 10),
    ],
)
def test_determine_nb_batches(
    descr: str,
    nb_rows_input_layer: int,
    nb_parallel: int,
    batchsize: int,
    cpu_count: int,
    bytes_usable: int,
    exp_nb_parallel: int,
    exp_nb_batches: int,
):
    # Set the parallelization parameters dependend on machine running the tests to keep
    # them stable regardless of the machine runninng the tests.
    config = _geoops_gpd.ParallelizationConfig(
        bytes_usable=bytes_usable, cpu_count=cpu_count
    )

    try:
        # Enable debug logging so debug code is covered
        logging.getLogger().setLevel(logging.DEBUG)

        res_nb_parallel, res_nb_batches = _geoops_gpd._determine_nb_batches(
            nb_rows_total=nb_rows_input_layer,
            nb_parallel=nb_parallel,
            batchsize=batchsize,
            parallelization_config=config,
        )
    finally:
        # Remove all handlers associated with the root logger object again.
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

    assert exp_nb_parallel == res_nb_parallel
    assert exp_nb_batches == res_nb_batches



"""
Module containing some general utilities.
"""

import datetime
import logging
import os
from collections.abc import Iterable
from typing import Any, Optional

logger = logging.getLogger(__name__)


class MissingRuntimeDependencyError(Exception):
    """
    Exception raised when an unsupported SQL statement is passed.

    Attributes:
        message (str): Exception message
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


def align_casing(string_to_align: str, strings_to_align_to: Iterable) -> str:
    """
    Search a string case-insentive in a list of string to align its casing.

    If the string is not found in strings_to_align_to, a ValueError is thrown.

    Args:
        string_to_align (str): string to align the casing of to strings_to_align_to.
        strings_to_align_to (Iterable): strings to align the casing with.

    Raises:
        ValueError: the string was not found in strings_to_align_to.

    Returns:
        str: the aligned string.
    """
    return align_casing_list([string_to_align], strings_to_align_to)[0]


def align_casing_list(
    strings_to_align: list[str],
    strings_to_align_to: Iterable,
    raise_on_missing: bool = True,
) -> list[str]:
    """
    Search the string caseintensitive in a list of strings.

    Args:
        strings_to_align (List[str]): strings to align the casing of to
            strings_to_align_to.
        strings_to_align_to (Iterable): strings to align the casing with.
        raise_on_missing (bool, optional): if True, a ValueError is raised if a string
            in ``strings_to_align`` is not found in ``strings_to_align_to``. If False,
            the casing in ``strings_to_align`` is retained in the output.

    Raises:
        ValueError: a string in strings_to_align was nog found in strings_to_align_to.

    Returns:
        List[str]: the aligned list of strings.
    """
    strings_to_align_to_upper_dict = {
        string.upper(): string for string in strings_to_align_to
    }
    strings_aligned = []
    for string in strings_to_align:
        string_aligned = strings_to_align_to_upper_dict.get(string.upper())
        if string_aligned is not None:
            strings_aligned.append(string_aligned)
        elif raise_on_missing:
            raise ValueError(f"{string} not available in: {strings_to_align_to}")
        else:
            strings_aligned.append(string)

    return strings_aligned


def report_progress(
    start_time: datetime.datetime,
    nb_done: int,
    nb_todo: int,
    operation: Optional[str] = None,
    nb_parallel: int = 1,
):
    # If logging level not enabled for INFO, no progress reporting...
    if logger.isEnabledFor(logging.INFO) is False:
        return

    message = format_progress(
        start_time=start_time,
        nb_done=nb_done,
        nb_todo=nb_todo,
        operation=operation,
        nb_parallel=nb_parallel,
    )
    if message is not None:
        if nb_done >= nb_todo:
            message += "\n"
        print(f"\r{message}", end="", flush=True)


def format_progress(
    start_time: datetime.datetime,
    nb_done: int,
    nb_todo: int,
    operation: Optional[str] = None,
    nb_parallel: int = 1,
) -> Optional[str]:
    # Init
    time_passed = (datetime.datetime.now() - start_time).total_seconds()
    pct_progress = 100.0 - (nb_todo - nb_done) * 100 / nb_todo
    nb_todo_str = f"{nb_todo:n}"
    nb_decimal = len(nb_todo_str)

    # If we haven't really started yet, don't report time estimate yet
    if nb_done == 0:
        return (
            f" ?: ?: ? left, {operation} done on {nb_done:{nb_decimal}n} of "
            f"{nb_todo:{nb_decimal}n} ({pct_progress:3.2f}%)    "
        )
    else:
        pct_progress = 100.0 - (nb_todo - nb_done) * 100 / nb_todo
        if time_passed > 0:
            # Else, report progress properly...
            processed_per_hour = (nb_done / time_passed) * 3600
            # Correct the nb processed per hour if running parallel
            if nb_done < nb_parallel:
                processed_per_hour = round(processed_per_hour * nb_parallel / nb_done)
            hours_to_go = (int)((nb_todo - nb_done) / processed_per_hour)
            min_to_go = (int)((((nb_todo - nb_done) / processed_per_hour) % 1) * 60)
            secs_to_go = (int)(
                ((((nb_todo - nb_done) / processed_per_hour) % 1) * 3600) % 60
            )
            time_left_str = f"{hours_to_go:02d}:{min_to_go:02d}:{secs_to_go:02d}"
            nb_left_str = f"{nb_done:{nb_decimal}n} of {nb_todo:{nb_decimal}n}"
            pct_str = f"({pct_progress:3.2f}%)    "
        elif pct_progress >= 100:
            time_left_str = "00:00:00"
            nb_left_str = f"{nb_done:{nb_decimal}n} of {nb_todo:{nb_decimal}n}"
            pct_str = f"({pct_progress:3.2f}%)    "
        else:
            return None
        message = f"{time_left_str} left, {operation} done on {nb_left_str} {pct_str}"
        return message


def formatbytes(bytes: float):
    """
    Return the given bytes as a human friendly KB, MB, GB, or TB string.
    """
    bytes_float = float(bytes)
    KB = float(1024)
    MB = float(KB**2)  # 1,048,576
    GB = float(KB**3)  # 1,073,741,824
    TB = float(KB**4)  # 1,099,511,627,776

    if bytes_float < KB:
        return "{} {}".format(bytes_float, "Bytes" if bytes_float > 1 else "Byte")
    elif KB <= bytes_float < MB:
        return f"{bytes_float / KB:.2f} KB"
    elif MB <= bytes_float < GB:
        return f"{bytes_float / MB:.2f} MB"
    elif GB <= bytes_float < TB:
        return f"{bytes_float / GB:.2f} GB"
    elif TB <= bytes_float:
        return f"{bytes_float / TB:.2f} TB"


def prepare_for_serialize(data: dict) -> dict:
    prepared: dict[str, Any] = {}
    for key, value in data.items():
        if isinstance(value, (dict)):
            prepared[key] = prepare_for_serialize(value)
        elif isinstance(value, (list, tuple)):
            prepared[key] = value
        else:
            prepared[key] = str(value)

    return prepared


class TempEnv:
    """
    Context manager to temporarily set/change environment variables.

    Existing values for variables are backed up and reset when the scope is left,
    variables that didn't exist before are deleted again.

    Args:
        envs (Dict[str, Any]): dict with environment variables to set.
    """

    def __init__(self, envs: dict[str, Any]):
        self._envs_backup: dict[str, str] = {}
        self._envs = envs

    def __enter__(self):
        # Only if a name and value is specified...
        for name, value in self._envs.items():
            # If the environment variable is already defined, make backup
            if name in os.environ:
                self._envs_backup[name] = os.environ[name]

            # Set env variable to value
            os.environ[name] = str(value)

    def __exit__(self, type, value, traceback):
        # Set variables that were backed up back to original value
        for name, value in self._envs_backup.items():
            # Recover backed up value
            os.environ[name] = value
        # For variables without backup, remove them
        for name, value in self._envs.items():
            if name not in self._envs_backup:
                if name in os.environ:
                    del os.environ[name]



"""
Tests for functionalities in ogr_util.
"""

import pytest
import shapely.geometry as sh_geom
from packaging import version

import geofileops as gfo
from geofileops._compat import SPATIALITE_GTE_51
from tests import test_helper


@pytest.mark.parametrize(
    "descr, sql, version_needed",
    [
        (
            "1_geos_basic",
            "SELECT ST_Buffer(ST_GeomFromText('POINT (5 5)'), 5) AS geom",
            "5.0",
        ),
        (
            "2_geos_advanced",
            "SELECT ST_ConcaveHull(ST_GeomFromText('POINT (5 5)')) geom",
            "5.0",
        ),
        (
            "3_geos_3100",
            "SELECT GeosMakeValid(ST_GeomFromText('POINT (5 5)')) AS geom",
            "5.1",
        ),
        (
            "4_geos_3110",
            """
            SELECT HilbertCode(
                       ST_GeomFromText('POINT (5 5)'),
                       ST_GeomFromText('POLYGON ((0 0, 0 10, 10 10, 10 0, 0 0))'),
                       5
                   ) AS geom
            """,
            "5.1",
        ),
    ],
)
def test_geos_functions(descr, sql, version_needed):
    """Test some geos functions from different spatialite versions."""
    test_path = test_helper.get_testfile(testfile="polygon-parcel")

    if version.parse(version_needed) < version.parse("5.1") or (
        version.parse(version_needed) >= version.parse("5.1") and SPATIALITE_GTE_51
    ):
        gfo.read_file(test_path, sql_stmt=sql)
    else:
        with pytest.raises(Exception, match="no such function"):
            gfo.read_file(test_path, sql_stmt=sql)


@pytest.mark.skipif(
    not test_helper.RUNS_LOCAL,
    reason="Don't this run on CI: just to document this behaviour.",
)
def test_st_difference_null(tmp_path):
    """
    ST_difference returns NULL when 2nd argument is NULL.

    This is the normal treatment of NULL values: `SELECT (1-NULL)` also results in NULL.

    In several spatial operations IIF statements are used to handle this situation.
    """
    input_path = test_helper.get_testfile(testfile="polygon-parcel")

    sql_stmt = """
        SELECT *
          FROM "{input_layer}"
         WHERE ST_difference({geometrycolumn}, NULL) IS NULL
    """
    output_path = tmp_path / "output.gpkg"
    gfo.select(input_path, output_path, sql_stmt=sql_stmt)

    input_gdf = gfo.read_file(input_path)
    output_gdf = gfo.read_file(output_path)
    assert len(input_gdf) == len(output_gdf)


@pytest.mark.skipif(
    not test_helper.RUNS_LOCAL,
    reason="Don't this run on CI: just to document this behaviour.",
)
@pytest.mark.parametrize("ogr_type", ["WKB", "WKT"])
def test_nested_collections(ogr_type):
    """
    A multi-geometry in a GeometryCollection is "flattened" by spatialite, which should
    not happen. For wkb's in some cases the result is even

    Bug report: https://groups.google.com/g/spatialite-users/c/GW-Ed0SL81Y
    """
    # Prepare test data
    input_path = test_helper.get_testfile(testfile="polygon-parcel")
    collection = sh_geom.GeometryCollection(
        [test_helper.TestData.multipolygon, test_helper.TestData.polygon_no_islands]
    )

    # Test
    if ogr_type == "WKT":
        sql_stmt = f"SELECT ST_GeomFromText('{collection.wkt}')"
    elif ogr_type == "WKB":
        sql_stmt = f"SELECT ST_GeomFromWKB(x'{collection.wkb_hex}')"

    test_gdf = gfo.read_file(input_path, sql_stmt=sql_stmt)
    if ogr_type == "WKT":
        # Due to a bug in spatialite this isn't equal even though it should be equal...
        assert test_gdf.geometry[0].wkt != collection.wkt
    elif ogr_type == "WKB":
        # Due to a worse bug in spatialite the wkb version is parsed to a POINT.
        assert test_gdf.geometry[0].wkt == "GEOMETRYCOLLECTION (POINT (0 0))"



"""
Module with utilities to format sql statements meant for use with ogr.
"""

from collections.abc import Iterable
from typing import Optional


class ColumnFormatter:
    """
    Format strings with the columns for use in sql statements.

    There are some specific hacks that are related to specific behaviours of ogr, mainly
    regarding the handling of the special "fid" column.
    """

    _aliases_cache: Optional[list[str]] = None

    def __init__(
        self,
        columns_asked: Optional[list[str]],
        columns_in_layer: Iterable[str],
        fid_column: str,
        table_alias: str = "",
        column_alias_prefix: str = "",
    ):
        """
        Format strings with column names for use in sql statements.

        Args:
            columns_asked (Optional[List[str]]): the column names to read from the
                file. If None, all available columns in the layer should be read.
                In addition to standard columns, it is also possible to specify "fid",
                a unique index available in all input files.
                Note that the "fid" will be aliased even if column_alias_prefix is "",
                eg. to "fid_1".
            columns_in_layer (Iterable[str]): the column names of the columns available
                in the layer that is being read from.
            fid_column (str): fid column name as reported by the gdal GetFIDColumn()
                function. For file types that don't store the fid this is "".
            table_alias (str, optional): table alias to be used.
                Defaults to "": no table alias.
            column_alias_prefix (str, optional): prefix to use for column aliases.
                Defaults to "": no prefix.

        Raises:
            ValueError: if columns are asked that are not available in the layer.
        """
        self._columns_in_layer = columns_in_layer
        self._fid_column = fid_column
        self._table_prefix = f"{table_alias}." if table_alias != "" else ""
        self._table_alias = table_alias
        self._columnname_prefix = column_alias_prefix

        # Now prepare the actual column list to use
        if columns_asked is not None:
            # Add special column "fid" to available columns so it can be specified
            columns_in_layer = list(columns_in_layer) + ["fid"]
            # Case-insensitive check if input1_columns contains columns not in layer...
            columns_in_layer_upper = {
                column.upper(): column for column in columns_in_layer
            }
            missing_columns = [
                col
                for col in columns_asked
                if (col.upper() not in columns_in_layer_upper)
            ]
            if len(missing_columns) > 0:
                raise ValueError(
                    "columns_asked contains columns not in columns_in_layer: "
                    f"{missing_columns}. Available: {columns_in_layer}"
                )

            # Create column list to keep in the casing of the original columns
            columns = [columns_in_layer_upper[col.upper()] for col in columns_asked]
        else:
            columns = list(columns_in_layer)
            columns_asked = columns

        self._columns = columns
        self._columns_asked = columns_asked

    def _columns_prefixed(self) -> list[str]:
        columns_prefixed = [
            f'{self._table_prefix}"{column}"' for column in self._columns
        ]
        columns_prefixed = self._fix_fid_columns(columns_prefixed)
        return columns_prefixed

    def _fix_fid_columns(self, columns: list[str]) -> list[str]:
        """
        Fix the fid columns.

        Useful if the "fid" special column needs some extra treatment:
            - if the fid_column name as reported by gdal is "", this means that the file
              format doesn't actually save the fid (eg. shapefile) but uses a row number
              in the file. When using sql in sql_dialect "SQLITE", "rowid" is the
              equivalent.
            - if the fid_column name as reported by gdal is "fid", ogr will treat the
              selected column as "fid" in the destination file as well, even if an alias
              is specified. To stop ogr from doing this, putting a CAST(... AS INT)
              around it prevents ogr from recognizing it.
              Remark: the ogr2ogr -unsetFid switch didn't help.
            - if the fid_column name as reported by gdal is some other string, "fid"
              just needs to be replaced by the fid_column value.
        """
        columns = list(columns)
        fid_column_indexes = [
            idx for idx, col in enumerate(self._columns) if col.upper() == "FID"
        ]
        if self._fid_column.lower() == "fid":
            # Put CAST() around "fid"
            for fid_column_index in fid_column_indexes:
                columns[fid_column_index] = f"CAST({columns[fid_column_index]} AS INT)"
        else:
            # Replace "fid" by the fid_column or rowid
            replace_fid_column = self._fid_column if self._fid_column != "" else "rowid"
            for fid_column_index in fid_column_indexes:
                columns[fid_column_index] = columns[fid_column_index].replace(
                    self._columns[fid_column_index], replace_fid_column
                )

        return columns

    def _aliases(self) -> list[str]:
        if self._aliases_cache is not None:
            return self._aliases_cache

        # Use columns_asked to keep asked casing
        aliases = [
            f"{self._columnname_prefix}{column}" for column in self._columns_asked
        ]

        # If no prefix, create a unique alias for fid column(s)
        if self._columnname_prefix == "":
            for alias_idx, alias in enumerate(aliases):
                if alias.lower() == "fid":
                    # If alias "fid", change it + make sure the alias isn't in use yet
                    for idx in range(1, 100):
                        alias_with_id = f"{alias}_{idx}"
                        if alias_with_id not in aliases:
                            aliases[alias_idx] = alias_with_id
                            break

        self._aliases_cache = aliases
        return self._aliases_cache

    def quoted(self) -> str:
        if len(self._columns) == 0:
            return ""

        columns_quoted = [f'"{column}"' for column in self._columns]
        columns_quoted = self._fix_fid_columns(columns_quoted)
        return f",{', '.join(columns_quoted)}"

    def prefixed(self) -> str:
        if len(self._columns) == 0:
            return ""
        return f",{', '.join(self._columns_prefixed())}"

    def prefixed_aliased(self):
        if len(self._columns) == 0:
            return ""

        columns_prefixed_aliased = [
            f'{column_prefixed} "{column_alias}"'
            for column_prefixed, column_alias in zip(
                self._columns_prefixed(), self._aliases()
            )
        ]
        return f",{', '.join(columns_prefixed_aliased)}"

    def null_aliased(self):
        if len(self._columns) == 0:
            return ""

        columns_null_aliased = [f'NULL "{alias}"' for alias in self._aliases()]
        return f",{', '.join(columns_null_aliased)}"

    def from_subselect(self, subselect_alias: str = "sub"):
        if len(self._columns) == 0:
            return ""

        prefix = "" if subselect_alias == "" else f"{subselect_alias}."
        columns_from_subselect = [f'{prefix}"{alias}"' for alias in self._aliases()]
        return f",{', '.join(columns_from_subselect)}"


def columns_quoted(columns: list[str]):
    if len(columns) == 0:
        return ""
    columns_quoted = [f'"{column}"' for column in columns]
    return f",{', '.join(columns_quoted)}"



"""
Tests for functionalities in _configoptions_helper.
"""

import os

import pytest

import geofileops as gfo
from geofileops.helpers import _configoptions_helper
from geofileops.helpers._configoptions_helper import ConfigOptions


@pytest.mark.parametrize(
    "value, default, expected",
    [
        ("TruE", False, True),
        ("YeS", False, True),
        ("1", False, True),
        ("nO", True, False),
        ("FalsE", True, False),
        ("0", True, False),
        ("", True, True),
        ("", False, False),
        (None, True, True),
        (None, False, False),
    ],
)
def test_get_bool(value, default, expected):
    test_key = "GFO_TEST_BOOL"
    if value is None:
        if test_key in os.environ:
            del os.environ[test_key]
    else:
        os.environ[test_key] = value

    result = _configoptions_helper.get_bool(test_key, default=default)
    if test_key in os.environ:
        del os.environ[test_key]

    assert result is expected


def test_get_bool_invalidvalue():
    test_key = "GFO_TEST_BOOL"
    with gfo.TempEnv({test_key: "INVALID"}):
        with pytest.raises(ValueError, match="invalid value for bool configoption"):
            _ = _configoptions_helper.get_bool(test_key, default="")


@pytest.mark.parametrize(
    "value, expected",
    [("TRUE", True), ("FALSE", False), (None, True)],
)
def test_remove_temp_files(value, expected):
    test_key = "GFO_REMOVE_TEMP_FILES"
    if value is None:
        if test_key in os.environ:
            del os.environ[test_key]
    else:
        os.environ[test_key] = value

    result = ConfigOptions.remove_temp_files
    if test_key in os.environ:
        del os.environ[test_key]

    assert result is expected
